/*
 * (C) Copyright 2019 Leo C. <erbl259-lmu@yahoo.de>
 *
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <stdio.h>
#include <stdlib.h>
#include "getopt.h"

# define EXIT_SUCCESS 0
# define EXIT_FAILURE 1

#define CPM_RS 128              /* CP/M record size */
#define NCOLS 4                 /* Default number of output columns */

/* The name this program was run with. */
char *program_name = "CVTSYM";

const char *sysfilename = "cpm3.sys";
const char *symfilename = "bnkbios3.sym";
const char *outfilename = "bnkbios3.asy";
int ncols = NCOLS;

unsigned int pseg, pseglen;
unsigned int cseg, cseglen;
int tot_symbols;
int pseg_symbols;
int cseg_symbols;

void usage (int status)
{
    if (status != EXIT_SUCCESS)
        fprintf (stderr, "Try '%s -h' for more information.\n", program_name);
    else {
        printf ("\
Usage: %s [options] [file]\n",
                program_name);
        fputs ("\
Convert symbol values from file (default: bnkbios3.sym).\n\
Options:\n\
", stdout);

        fputs ("\
  -b <file>         Get segment addresses from <file> (default: cpm3.sys)\n\
  -n number         Number of output columns (1...8)\n\
  -o <file>         Place output into <file> (default: bnkbios3.asy)\n\
", stdout);

        fputs ("\
  -h                display this help and exit\n"
#if 0
"  -v                output version information and exit\n"
#endif
, stdout);

    }
    exit (status);
}

static void
ps(const char *s)
{
    while(*s)
        putc(*s++, stderr);
}

void error(const char *msg, const char *arg)
{
    ps(program_name);
    ps(": ");

    ps(msg);
    if (arg != NULL)
        ps(arg);
    putc('\n', stderr);

    exit (EXIT_FAILURE);
}

void get_segments(void)
{
    int i,n;
    register FILE *sysf;
    unsigned int seg, seglen;

    if (!(sysf = fopen(sysfilename, "rb")))
        error("File not found: ", sysfilename);

    /* skip first record */
    fseek(sysf, CPM_RS, 0);

    n = 0; i = 0;
    while (n < 2 && i < 4) {
        i++;
        if (fscanf(sysf, " BNKBIOS3 SPR%4x%4x", &seg, &seglen) == 2) {
            ++n;
            if (n == 1) {
                pseg = seg;
                pseglen = seglen;
            } else {
                cseg = seg;
                cseglen = seglen;
            }
        }
    }

    fclose(sysf);
    if (n != 2)
        error("Segments not found", NULL);
}

void convert_symbols(void)
{
    int i;
    unsigned val;
    char sym[9];
    FILE *symf;
    FILE *outf;

    if (!(symf = fopen(symfilename, "r")))
        error("File not found: ", symfilename);
    if (!(outf = fopen(outfilename, "wb")))
        error("Can't open: ", symfilename);

    i = ncols;
    while (fscanf(symf, "%4x%8s", &val, sym) == 2) {
        ++tot_symbols;
        if (val < pseglen) {
            ++pseg_symbols;
            val += pseg;
        } else {
            if (val < pseglen+cseglen) {
                ++cseg_symbols;
                val += cseg-pseglen;
            }
        }

#ifdef HI_TECH_C
        fprintf(outf, "%4.4x %s", val, sym);
#else
        fprintf(outf, "%4.4X %s", val, sym);
#endif
        if (--i == 0) {
            fputs("\r\n", outf);
            i = ncols;
        } else {
            fputc('\t', outf);
        }
    }

    fputs("\r\n", outf);
    fputc(0x1a, outf);
    if (fclose(outf) != 0)
        error("Error writing: ", outfilename);
}

int main(int argc, char **argv)
{

    int optc;

#ifndef CPM
    program_name = argv[0];
#else
    argv[0] = program_name;
#endif

    /* opterr = 0; */
    while ((optc = getopt (argc, argv, "B:b:N:n:O:o:Hh")) != EOF) {

        switch(optc) {
        case 'b':
        case 'B':
            sysfilename = optarg;
            break;
        case 'o':
        case 'O':
            outfilename = optarg;
            break;
        case 'n':
        case 'N':
            ncols = atoi(optarg);
            if (ncols < 1 || ncols > 8)
                error("invalid value for option n", NULL);
            break;

        case 'h':
        case 'H':
            usage (EXIT_SUCCESS);
            break;
        default:
            /* '?' */
            exit(EXIT_FAILURE);
            break;
        }
    }

    if (optind < argc)
        symfilename = argv[optind];



    get_segments();
#if DEBUG
    printf("%x   %x     %x   %x\n", pseg, pseglen, cseg, cseglen);
#endif

    convert_symbols();

    printf("Symbols total: %d    PSEG: %d    CSEG %d    other %d\n",
        tot_symbols, pseg_symbols, cseg_symbols, tot_symbols - pseg_symbols - cseg_symbols);

    exit(0);

}
